/*
 * Sept 7, 2020 Success! The trick seems to be that if the write succeeds,
 * even though the Arduino is supposedly sending back what you wrote, it
 * doesn't the first few times. So keep writing until it does.
 * I may have to put together a protocol like that -- keep writing until
 * I get an ACK of some sort.
*/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/serial.h>

#if 1
/*
 * This came from:
 * https://github.com/jdourlens/FirmataC/blob/master/src/serial.c
 */
typedef struct		s_serial
{
  int			port_is_open;
  char			*port_name;
  int			baud_rate;
  char			*error_msg;

  int			port_fd;
  struct termios	settings_orig;
  struct termios	settings;
  int			tx;
  int			rx;
}			t_serial;

int serial_setBaud(t_serial *serial, int baud)
{
  speed_t spd;
  switch (baud) {
  case 230400:    spd = B230400;  break;
  case 115200:    spd = B115200;  break;
  case 57600:     spd = B57600;   break;
  case 38400:     spd = B38400;   break;
  case 19200:     spd = B19200;   break;
  case 9600:      spd = B9600;    break;
  case 4800:      spd = B4800;    break;
  case 2400:      spd = B2400;    break;
  case 1800:      spd = B1800;    break;
  case 1200:      spd = B1200;    break;
  case 600:       spd = B600;     break;
  case 300:       spd = B300;     break;
  case 200:       spd = B200;     break;
  case 150:       spd = B150;     break;
  case 134:       spd = B134;     break;
  case 110:       spd = B110;     break;
  case 75:        spd = B75;      break;
  case 50:        spd = B50;      break;
#ifdef B460800
  case 460800:    spd = B460800;  break;
#endif
#ifdef B500000
  case 500000:    spd = B500000;  break;
#endif
#ifdef B576000
  case 576000:    spd = B576000;  break;
#endif
#ifdef B921600
  case 921600:    spd = B921600;  break;
#endif
#ifdef B1000000
  case 1000000:   spd = B1000000; break;
#endif
#ifdef B1152000
  case 1152000:   spd = B1152000; break;
#endif
#ifdef B1500000
  case 1500000:   spd = B1500000; break;
#endif
#ifdef B2000000
  case 2000000:   spd = B2000000; break;
#endif
#ifdef B2500000
  case 2500000:   spd = B2500000; break;
#endif
#ifdef B3000000
  case 3000000:   spd = B3000000; break;
#endif
#ifdef B3500000
  case 3500000:   spd = B3500000; break;
#endif
#ifdef B4000000
  case 4000000:   spd = B4000000; break;
#endif
#ifdef B7200
  case 7200:      spd = B7200;    break;
#endif
#ifdef B14400
  case 14400:     spd = B14400;   break;
#endif
#ifdef B28800
  case 28800:     spd = B28800;   break;
#endif
#ifdef B76800
  case 76800:     spd = B76800;   break;
#endif
  default: {
    return -1;
  }
  }
  cfsetospeed(&(serial->settings), spd);
  cfsetispeed(&(serial->settings), spd);
  if (tcsetattr(serial->port_fd, TCSANOW, &(serial->settings)) < 0)
    return (-1);
  return (0);
}

t_serial *serial_new()
{
  t_serial  *res;

  res = malloc(sizeof(t_serial));
  if (!res)
    {
      return (NULL);
    }
  res->port_is_open = 0;
  res->baud_rate = 38400;
  res->tx = 0;
  res->rx = 0;
  return (res);
}

int		serial_open(t_serial *serial, char *name)
{
  struct serial_struct	kernel_serial_settings;
  int			bits;

  serial->port_fd = open(name, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (serial->port_fd < 0)
    {
      if (errno == EACCES)
	{
	  perror("Unable to access, insufficient permission");
	}
      else if (errno == EISDIR)
	{
	  perror("Unable to open, Object is a directory, not a serial port");
	}
      else if (errno == ENODEV || errno == ENXIO)
	{
	  perror("Unable to open, Serial port hardware not installed");
	}
      else if (errno == ENOENT)
	{
	  perror("Unable to open, Device name does not exist");
	}
      else
	{
	  perror("Unable to open: Unknown error.");
	}
      return (-1);
    }
  if (ioctl(serial->port_fd, TIOCMGET, &bits) < 0)
    {
      close(serial->port_fd);
      perror("Unable to query serial port signals");
      return (-1);
    }
  bits &= ~(TIOCM_DTR | TIOCM_RTS);
  if (ioctl(serial->port_fd, TIOCMSET, &bits) < 0)
    {
      close(serial->port_fd);
      perror("Unable to control serial port signals");
      return (-1);
    }
  if (tcgetattr(serial->port_fd, &(serial->settings_orig)) != 0)
    {
      close(serial->port_fd);
      perror("Unable to query serial port settings (perhaps not a serial port)");
      return (-1);
    }
  memset(&(serial->settings), 0, sizeof(struct termios));
  serial->settings.c_iflag = IGNBRK | IGNPAR;
  serial->settings.c_cflag = CS8 | CREAD | HUPCL | CLOCAL;
  serial_setBaud(serial, serial->baud_rate);
  if (ioctl(serial->port_fd, TIOCGSERIAL, &kernel_serial_settings) == 0)
    {
      kernel_serial_settings.flags |= ASYNC_LOW_LATENCY;
      ioctl(serial->port_fd, TIOCSSERIAL, &kernel_serial_settings);
    }
  tcflush(serial->port_fd, TCIFLUSH);
  serial->port_name = name;
  serial->port_is_open = 1;
  return 0;
}
#endif
#if 0
/*
 * A lot of the following settings came from:
 * https://github.com/jdourlens/FirmataC/blob/master/src/serial.c
 */
int serial_open(char *port, int baud)
{
	int fd, bits;
	struct termios term;
	struct serial_struct kernel_serial_settings;

	if ((fd = open(port, O_RDWR|O_NOCTTY|O_NONBLOCK)) == -1) {
		perror("open()");
		return -1;
	}
	if (ioctl(fd, TIOCMGET, &bits) == -1) {
		perror("ioctl(TIOCMGET)");
		return -1;
	}
	bits &= ~(TIOCM_DTR | TIOCM_RTS);
	if (ioctl(fd, TIOCMSET, &bits) == -1) {
		perror("ioctl(TIOCMSET)");
		return -1;
	}
	//tcgetattr(fd, &term);
	memset(&term, 0, sizeof(term));
	term.c_iflag = IGNBRK | IGNPAR;
	term.c_cflag = CS8 | CREAD | HUPCL | CLOCAL;
	cfsetispeed(&term, B9600);
	cfsetospeed(&term, B9600);
	tcsetattr(fd, TCSANOW, &term);
	if (ioctl(fd, TIOCGSERIAL, &kernel_serial_settings) == 0) {
		kernel_serial_settings.flags != ASYNC_LOW_LATENCY;
		ioctl(fd, TIOCSSERIAL, &kernel_serial_settings);
	}
	tcflush(fd, TCIFLUSH);
	return fd;
}
#endif
#if 0
/*
 * This is my own adapted verison.
 * Much of this was done by reading through:
 * https://www.cmrr.umn.edu/~strupp/serial.html
 */
int serial_open(char *port, int baud)
{
	int fd, bits;
	struct termios term;
	struct serial_struct kernel_serial_settings;

	if ((fd = open(port, O_RDWR|O_NOCTTY|O_NONBLOCK)) == -1) {
		perror("open()");
		return -1;
	}
	if (ioctl(fd, TIOCMGET, &bits) == -1) {
		perror("ioctl(TIOCMGET)");
		return -1;
	}
	bits &= ~(TIOCM_DTR | TIOCM_RTS);
	if (ioctl(fd, TIOCMSET, &bits) == -1) {
		perror("ioctl(TIOCMSET)");
		return -1;
	}
	tcgetattr(fd, &term);
	//memset(&term, 0, sizeof(term));

	// enable the receiver and set local mode
   term.c_cflag |= (CLOCAL | CREAD);

   // the following gives 8N1
   term.c_cflag &= ~PARENB;
   term.c_cflag &= ~CSTOPB;
   term.c_cflag &= ~CSIZE;
	term.c_cflag |= CS8;

   // set to raw input
	term.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

   // no software flow control
   term.c_iflag &= ~(IXON | IXOFF | IXANY);

	// raw output (no output processing)
   term.c_oflag &= ~OPOST;

	cfsetispeed(&term, B9600);
	cfsetospeed(&term, B9600);

	tcsetattr(fd, TCSANOW, &term);
/*
	if (ioctl(fd, TIOCGSERIAL, &kernel_serial_settings) == 0) {
		kernel_serial_settings.flags != ASYNC_LOW_LATENCY;
		ioctl(fd, TIOCSSERIAL, &kernel_serial_settings);
	}
*/
	tcflush(fd, TCIFLUSH);
	return fd;
}
#endif

int main(int argc, char **argv)
{
	char *serialport = "/dev/ttyACM0";
	char *data = "Hello from Raspberry Pi!\n";
	int fd;
	ssize_t n;

	t_serial *ser;
	if ((ser = serial_new()) == NULL) {
		perror("serial_new()");
		exit(EXIT_FAILURE);
	}
	ser->baud_rate = 9600;
	if (serial_open(ser, serialport) == -1) {
		fprintf(stderr, "open(%s) failed: %s\n", serialport, strerror(errno));
		exit(EXIT_FAILURE);
	}
	fd = ser->port_fd;
/*
	if ((fd = serial_open(serialport, 9600)) == -1) {
		fprintf(stderr, "open(%s) failed: %s\n", serialport, strerror(errno));
		exit(EXIT_FAILURE);
	}
*/
/*
	if ((fd = open(serialport, O_RDWR|O_NONBLOCK)) == -1) {
		perror("open()");
		return -1;
	}
*/
tcdrain(fd);
while (1) {
	if ((n = write(fd, data, strlen(data))) == -1) {
		fprintf(stderr, "write() failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	printf("Wrote %ld bytes to %s: %s\n", n, serialport, data);
	tcdrain(fd);

	char buf[100];
	memset(buf, 0, 100);
	if ((n = read(fd, buf, 100)) == -1) {
		if (errno != EAGAIN) {
			fprintf(stderr, "read() failed: (%d) %s\n", errno, strerror(errno));
			exit(EXIT_SUCCESS);
		}
		printf("read() returned -1 and EAGAIN\n");
	}
	printf("Read returned %ld bytes.\n", n);
	if (n > 100)
		printf("READ over 100 bytes !!!!\n");
	else if (n > 0) {
		printf("Char/byte were: %s\n", buf);
/*
	int b;
	   for (b = 0; b < n; b++) {
			printf(" ");
			if (isascii(buf[b]) && buf[b] != 0x0d && buf[b] != 0x0a)
				printf("%c,", buf[b]);
			printf("0x%x", buf[b]);
		}
		printf("\n");
*/
	}
	sleep(1);
}

	close(fd);

	return EXIT_SUCCESS;
}
