/*
 * brview - Brain viewer for izzy.
 *
 * Note that this uses GTK 3 because GTK 4 is not available for the Ubuntu
 * tegra distribution that came with my Jetson TX1 (it seems).
*/

#include <errno.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gtk/gtk.h>
#include "brview.h"

static GdkPixbuf *pb;
static GtkWidget *image;

/*
 * The test data below is 32-bits per pixel RGB: 8-unused, 8-R, 8-G, 8-B.
 * Width = 8 pixels, Height = 9 pixels
*/
#define IWIDTH	8
#define IHEIGHT	9
static unsigned char data1[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static unsigned char data2[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

typedef struct {
	brview_t *bv;
	int argc;
	char **argv;
	sem_t sem;
} bvthread_data_t;

static void brview_errno(char *text, int errnoval)
{
    fprintf(stderr, "brview: %s - %s\n", text, strerror(errnoval));
}

static void brview_errval(char *text, int errnoval)
{
    fprintf(stderr, "brview: %s - %d\n", text, errnoval);
}

static void brview_errstr(char *text)
{
    fprintf(stderr, "brview: %s\n", text);
}

enum bv_action {
	BV_ACTION_IMAGE_CREATE,
	BV_ACTION_IMAGE_UPDATE
};

typedef struct {
	enum bv_action	action; // must be the first member
	brview_t		*bv;
	brview_image_t	*img;
} bv_action_image_create_t;

typedef struct {
	enum bv_action	action; // must be the first member
	brview_image_t	*img;
} bv_action_image_update_t;

static gboolean img_draw_callback(GtkWidget *widget, cairo_t *cr, gpointer data)
{
    brview_image_t *img = data;

#if 0
	cairo_surface_flush(img->surface);
	unsigned char *isurf = cairo_image_surface_get_data(img->surface);
	printf("width=%d, height=%d, stride=%d\n",
		img->width, img->height, img->stride);
	for (int y = 0; y < img->height; y++) {
		for (int x = 0; x < img->stride; x++)
			printf("0x%x ", (int) isurf[x+(y*img->stride)]);
		printf("\n");
	}
#endif

    cairo_set_source_surface(cr, img->surface, 0, 0);
    cairo_paint(cr);

    return FALSE;
}

static int bv_action_image_create(bv_action_image_create_t *a)
{
	brview_image_t *img = a->img;

#if 0
	a->data = data1;
	a->width = IWIDTH;
	a->height = IHEIGHT;
	a->stride = brview_image_get_stride(a->width);
#endif

	if ((img->drawing_area = gtk_drawing_area_new()) == NULL) {
		brview_errstr("gtk_drawing_area_new failed");
		return -1;
	}
	gtk_fixed_put(GTK_FIXED(a->bv->fixed), img->drawing_area, img->x, img->y);
	gtk_widget_set_size_request(img->drawing_area, img->width, img->height);
	g_signal_connect(G_OBJECT (img->drawing_area), "draw",
						G_CALLBACK (img_draw_callback), img);

	// GOBACK: not sure if the following is the correct way to flush
	gtk_widget_show_all (a->bv->window);

	free(a);

	return -1;
}

static void bv_action_image_update(bv_action_image_update_t *a)
{
	brview_image_t *img = a->img;

	cairo_surface_flush(img->surface);
	//isurf = cairo_image_surface_get_data(img->surface);
	cairo_surface_mark_dirty(img->surface);
	gtk_widget_queue_draw(img->drawing_area); // tell it to redraw
}

/*
 * idle_callback - Called from the GTK main loop when things are idle.
 * We use this because we have a GTK thread and then the thread(s) which
 * is the main program. To have GTK stuff be done only from the GTK thread,
 * whenever the main program wants something done, it adds this callback
 * as a one-shot-thing, and this callback is then called by the GTK thread
 * to do the work.
 * Info: https://developer.gimp.org/api/2.0/gdk/gdk-Threads.html
*/
static gboolean idle_callback(gpointer gdata)
{
	enum bv_action *action = gdata;

	switch (*action) {
	case BV_ACTION_IMAGE_CREATE:
		printf("in idle_callback, action=BV_ACTION_IMAGE_CREATE\n");
		bv_action_image_create(gdata);
		break;
	case BV_ACTION_IMAGE_UPDATE:
		printf("in idle_callback, action=BV_ACTION_IMAGE_UPDATE\n");
		bv_action_image_update(gdata);
		break;
	default:
		brview_errno("idle_callback", EINVAL);
		break;
	}

	return FALSE; // remove the callback so that it's called only once
}

/*
 * brview_image_get_stride - Get the number of bytes required for a row
 * of an image of a given width. The only supported format is
 * RGB with 4 bytes per pixel with 3 bytes for the RGB and the upper
 * byte unused. The stride returned may actually be longer than the
 * width*4 for high-performance rendering.
 * width: The width of each row in the image in pixels.
 * Returns:
 * The stride are described above.
*/
int brview_image_get_stride(int width)
{
	return cairo_format_stride_for_width(CAIRO_FORMAT_RGB24, width);
}

/*
 * brview_image_create - Creates an image object. Each row must be
 * must be RGB with 4 bytes per pixel with 3 bytes for the 
 * RGB and the upper byte unused. The actual number of bytes per row
 * (the stride) may actually be longer than the width*4 for 
 * high-performance rendering. Call brview_image_get_stride() to find
 * out what stride to use for each row before allocating memory for the
 * image.
 * brview_t *bv: The handle returned by brview_setup().
 * char *img: A buffer containing an image.
 * int width, height: The dimensions in pixels.
 * Returns:
 * non-NULL: a pointer to an allocated brview_image_t structure
 * NULL: error, errno is set
*/
brview_image_t *brview_image_create(brview_t *bv, unsigned char *data, int x, int y, int width, int height, int stride)
{
	bv_action_image_create_t *a;
	brview_image_t *img;
	cairo_status_t cstatus;

	if ((a = malloc(sizeof(*a))) == NULL) {
		brview_errno("malloc", errno);
		return NULL;
	}
	a->action = BV_ACTION_IMAGE_CREATE;
	a->bv = bv;
	if ((img = malloc(sizeof(*img))) == NULL) {
		brview_errno("malloc", errno);
		return NULL;
	}
	a->img = img;
	img->data = data;
	img->x = x;
	img->y = y;
	img->width = width;
	img->height = height;
	img->stride = stride;
	img->surface = cairo_image_surface_create_for_data(
		img->data, CAIRO_FORMAT_RGB24,
		width, height, stride);
	if ((cstatus = cairo_surface_status(img->surface)) != CAIRO_STATUS_SUCCESS) {
		brview_errval("cairo_image_surface_create_for_data", cstatus);
		return NULL;
		
	}

	gdk_threads_add_idle(idle_callback, a);

	return img;
}

/*
 * Adapted from: https://www.cairographics.org/manual/cairo-Image-Surfaces.html
*/
int brview_image_update(brview_image_t *img)
{
	bv_action_image_update_t *a;

	if ((a = malloc(sizeof(*a))) == NULL) {
		brview_errno("malloc", errno);
		return -1;
	}
	a->action = BV_ACTION_IMAGE_UPDATE;
	a->img = img;
	gdk_threads_add_idle(idle_callback, a);

	return 0;
}

static void create_image(GtkWidget *window)
{

	pb = gdk_pixbuf_new_from_data(data1, GDK_COLORSPACE_RGB, 
		FALSE, 8, 8, 9, 3*8, NULL, NULL);
	image = gtk_image_new_from_pixbuf(pb);
	gtk_container_add (GTK_CONTAINER (window), image);
	
}

static void activate(GtkApplication *app, gpointer user_data)
{
	GtkWidget *window;
	GtkWidget *fixed;
	GtkWidget *button;
	GtkWidget *button_box;
	bvthread_data_t *td = user_data;

	window = gtk_application_window_new(app);
	gtk_window_set_title(GTK_WINDOW(window), "Brain Viewer");
	gtk_window_set_default_size(GTK_WINDOW(window), 200, 200);

    /*
     * A window can have only one child. One type of container is a fixed
     * type of container which can have multiple children at fixed locations.
     * For other types of layout containers, see:
     * https://zetcode.com/gui/gtk2/gtklayoutmanagement/
    */
    fixed = gtk_fixed_new();
    gtk_container_add(GTK_CONTAINER(window), fixed);

#if 0
	button_box = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
	gtk_container_add (GTK_CONTAINER (window), button_box);

	button = gtk_button_new_with_label ("Hello World");
	g_signal_connect (button, "clicked", G_CALLBACK (print_hello), NULL);
	g_signal_connect_swapped (button, "clicked", G_CALLBACK (gtk_widget_destroy), window);
	gtk_container_add (GTK_CONTAINER (button_box), button);
#endif

//	create_image(window);

	gtk_widget_show_all (window);

	td->bv->window = window;
	td->bv->fixed = fixed;

	sem_post(&td->sem);
}

static void *brview_thread(void *data)
{
	GtkApplication *app;
	int status;
	bvthread_data_t *td = data;

	app = gtk_application_new ("org.brview", G_APPLICATION_FLAGS_NONE);
	g_signal_connect (app, "activate", G_CALLBACK (activate), td);
	status = g_application_run (G_APPLICATION (app), td->argc, td->argv);
	g_object_unref (app);

	return (void *) (long int) status;
}

brview_t *brview_setup(int argc, char **argv)
{
	int status;
	brview_t *bv;
	bvthread_data_t *td;
	pthread_t tid;
	struct timespec tm;

	if ((td = malloc(sizeof(*td))) == NULL) {
		brview_errno("malloc", errno);
		return NULL;
	}
	if ((td->bv = malloc(sizeof(*(td->bv)))) == NULL) {
		free(td);
		brview_errno("malloc", errno);
		return NULL;
	}
	if (sem_init(&td->sem, 0, 0) == -1) {
		free(td->bv);
		free(td);
		brview_errno("sem_init", errno);
		return NULL;
	}
	td->argc = argc;
	td->argv = argv;
	if ((status = pthread_create(&tid, NULL, brview_thread, td)) != 0) {
		sem_destroy(&td->sem);
		free(td->bv);
		free(td);
		brview_errno("pthread_create", status);
		errno = status;
		return NULL;
	}
	clock_gettime(CLOCK_REALTIME, &tm);
	tm.tv_sec += 1;
	if (sem_timedwait(&td->sem, &tm) == -1) {
		sem_destroy(&td->sem);
		free(td->bv);
		free(td);
		brview_errno("sem_timedwait", errno);
		return NULL;
	}
	bv = td->bv;
	sem_destroy(&td->sem);
	free(td);
	return bv;
}
