#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "brview.h"
#include "obrec.h"

// TODO: The following should be in a common communication header used
// here and by darknet
#define COMM_BUFMAX_NEW_SHM		(NAME_MAX+10+5+5+1)
#define COMM_BUFMAX_UPDATED_IMG	(NAME_MAX+12+1)
#define COMM_BUFMAX_MAX			COMM_BUFMAX_NEW_SHM // the lagest one
#define COMM_BUFMAX_DETECTION   512

int obrec_connect(int *sockfd, unsigned char **shmptr, char *sockname, char *shmname, int w, int h)
{
	int shmfd, shmlen;
	unsigned char *tshmptr;
    int tsockfd;
    struct sockaddr_un saddr;
	char buf[COMM_BUFMAX_NEW_SHM];

	if ((shmfd = shm_open(shmname, O_RDWR|O_CREAT, S_IRUSR)) == -1) {
		perror("shm_open()");
		return -1;
	}
	shmlen = w*h*4;
printf("w=%d, h=%d\n", w, h);
	if (ftruncate(shmfd, shmlen) == -1) {
		perror("ftruncate()");
		close(shmfd);
		return -1;
	}
    if ((tshmptr = mmap(0, shmlen, PROT_READ|PROT_WRITE, MAP_SHARED, shmfd, 0)) == MAP_FAILED) {
		close(shmfd);
		return -1;
	}
	close(shmfd);

    // create an endpoint for the communication
    // domain=AF_LOCAL (synonym for AF_UNIX)
    // type=SOCK_STREAM for TCP
    // procotol=0 since only one protocol exists
    if ((tsockfd = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1) {
        perror("socket()");
		munmap(tshmptr, shmlen);
        return -1;
    }

    memset(&saddr, 0, sizeof(saddr));
    saddr.sun_family = AF_LOCAL;
    strncpy(saddr.sun_path, sockname, sizeof(saddr.sun_path)-1);
    if (connect(tsockfd, (struct sockaddr *) &saddr,
            sizeof(saddr)) == -1) {
        perror("connect()");
		munmap(tshmptr, shmlen);
		close(tsockfd);
        return -1;
    }

	*sockfd = tsockfd;
	*shmptr = tshmptr;

    snprintf(buf, sizeof(buf), "shm=%s:w=%d:h=%d", shmname, w, h);
	if (write(*sockfd, buf, strlen(buf)) == -1) {
		perror("Sending shared memory name to socket");
		return -1;
	}

	return 0;
}

int obrec_parse_objects_string(objectr_t **obrs, char *dbuf)
{
	char *c, *p;
	objectr_t tmp, *tobr;
	int cnt = 1;

	*obrs = NULL;

	c = dbuf;
	while (*c != '\0') {
		if ((p = strchr(c, ':')) == NULL)
			goto fmterrorout;
		*p = '\0';
		if (strlen(c) == 0)
			goto fmterrorout;
		if ((tmp.name = strdup(c)) == NULL)
			goto memerrorout;
		c = p + 1;
		
		if ((p = strchr(c, ':')) == NULL)
			goto fmterrorout;
		if (c+1 == p)
			goto fmterrorout;
		*p = '\0';
		tmp.prob = atof(c);
		c = p + 1;
		
		for (int i = 0; i < 4; i++) {
			if ((i < 3 && (p = strchr(c, ':')) == NULL) ||
				(i == 3 && (p = strchr(c, '\n')) == NULL))
				goto fmterrorout;
			if (c == p)
				goto fmterrorout;
			*p = '\0';
			switch (i) {
			case 0: tmp.left = atoi(c); break;
			case 1: tmp.right = atoi(c); break;
			case 2: tmp.top = atoi(c); break;
			case 3: tmp.bot = atoi(c); break;
			}
			c = p + 1;
		}

		cnt++;

		if ((tobr = calloc(1, sizeof(objectr_t))) == NULL)
			goto memerrorout;
		memcpy(tobr, &tmp, sizeof(objectr_t));
		if (*obrs == NULL) {
			tobr->next = NULL;
			*obrs = tobr;
		} else {
			tobr->next = *obrs;
			*obrs = tobr;
		}
	}
	return 0;

fmterrorout:
	fprintf(stderr, "Error in object %d in buffer at '%s'\n", cnt, c);
	return -1;

memerrorout:
	fprintf(stderr, "No memory\n");
	return -1;
}

int obrec_find_objects_in_image(objectr_t **objectrs, int sockfd, unsigned char *shmptr, char *shmname, brview_image_t *img)
{
	int x, y;
	int nbytes, off;
	char buf[COMM_BUFMAX_UPDATED_IMG], dbuf[COMM_BUFMAX_DETECTION];
	unsigned char *i, *s;
	objectr_t *tobr;

	*objectrs = NULL;

	// copy from our image to the memory shared between us and the
	// object recognition program
	s = shmptr;
	for (y = 0; y < img->height; y++) {
		i = &(img->data[y*img->stride]);
		for (x = 0; x < img->width; x++) {
			s[0] = i[2];
			s[1] = i[1];
			s[2] = i[0];
			i += 4;
			s += 3; 
		}
	}
    snprintf(buf, sizeof(buf), "shm=%s:updated", shmname);
	if (write(sockfd, buf, strlen(buf)) == -1) {
		perror("Sending notification to socket");
		return -1;
	}
	off = 0;
	do {
	    if ((nbytes = read(sockfd, &dbuf[off], sizeof(dbuf)-off)) == -1) {
			perror("Reading detection from socket");
			return -1;
		}
		if (nbytes == 0)
			break;
		off += nbytes;
	} while (off < sizeof(dbuf) && dbuf[off-1] != '\0');
	printf("dbuf[%d]=%s\n", off-1, dbuf[off-1]=='\0' ? "NULL" : "non-NULL");
	if (off > 1 && dbuf[off-1] != '\0') {
		fprintf(stderr, "Too many objects, room in buffer for only %ld bytes\n", sizeof(dbuf));
		return -1;
	} 

	// copy back to our image to the memory shared between us and the
	// object recognition program, it now has rectangles drawn around
	// the recognized objects along with their names
	s = shmptr;
	for (y = 0; y < img->height; y++) {
		i = &(img->data[y*img->stride]);
		for (x = 0; x < img->width; x++) {
			i[2] = s[0];
			i[1] = s[1];
			i[0] = s[2];
			i += 4;
			s += 3; 
		}
	}

    printf("Got detection(s): %s\n", dbuf);
	if (obrec_parse_objects_string(objectrs, dbuf) == -1)
		return -1;
	tobr = *objectrs;
	while (tobr != NULL) {
		printf("name: %s, probability: %.2f, left: %d, right: %d, top: %d, bot: %d\n",
			tobr->name, tobr->prob, tobr->left, tobr->right, tobr->top, tobr->bot);
		tobr = tobr->next;
	}

	return 0;
}

void obrec_free_objectr_list(objectr_t *obrs)
{
	objectr_t *tobr;

	while (obrs != NULL) {
		free(obrs->name);
		tobr = obrs;
		obrs = obrs->next;
		free(tobr);
	}
}
